from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

from .core import register


@register('sp.fullpath')
def get_full_path(path: str) -> str:
    """Resolve all implicit and relative path components
    to give an absolute path to a file or directory"""
    return str(Path(path).resolve().absolute())


@register('sp.timestamp')
def get_timestamp(fmt: Optional[str] = None) -> str:
    """Returns a timestamp in the format provided; if not provided, use
    year-month-day_hour-minute-second."""

    fmt = fmt or "%Y-%m-%d_%H-%M-%S"
    return datetime.now(tz=timezone.utc).strftime(fmt)



import copy
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass, field, is_dataclass  # noqa: F401
from pathlib import Path
from typing import (Any, Callable, Dict, Iterator, Optional, Protocol,
                    Sequence, Type, TypeVar, Union)

from omegaconf import MISSING, DictConfig, OmegaConf, open_dict

# for return type
RT = TypeVar('RT')


class DataClass(Protocol):
    # from https://stackoverflow.com/a/55240861
    __dataclass_fields__: Dict[str, Any]

    def __instancecheck__(self, __instance: Any) -> bool:
        return is_dataclass(__instance)


ConfigType = Union[DictConfig, Dict[str, Any], Type[DataClass], str]


@dataclass
class ParamSpec:
    name: str
    path: str
    value: Any
    node: Optional[DictConfig]


def traverse(config_node: ConfigType) -> Iterator[ParamSpec]:
    """Returns all keys for a config node"""

    config_node = cast(config_node)

    for key in config_node.keys():
        if OmegaConf.is_missing(config_node, key):
            value = MISSING
        elif OmegaConf.is_interpolation(config_node, str(key)):
            value = OmegaConf.to_container(config_node)[key]   # type: ignore
        else:
            value = config_node[key]

        if isinstance(value, DictConfig):
            for p in traverse(value):
                yield ParamSpec(name=p.name,
                                path=f'{key}.{p.path}',
                                value=p.value,
                                node=p.node)
        else:
            yield ParamSpec(name=str(key),
                            path=str(key),
                            value=value,
                            node=config_node)


def validate(config_node: ConfigType) -> DictConfig:
    """Check if all attributes are resolve and not missing"""

    if not isinstance(config_node, DictConfig):
        raise TypeError(f'`{config_node}` is not a DictConfig!')

    for spec in traverse(config_node):
        if OmegaConf.is_missing(spec.node, spec.name):
            raise ValueError(f'Missing value for `{spec.path}`')
        if OmegaConf.is_interpolation(spec.node, spec.name):
            try:
                getattr(spec.node, spec.name)
            except Exception:
                raise ValueError(f'Interpolation for `{spec.path}` '
                                 'not resolved')

    config_node = copy.deepcopy(config_node)
    OmegaConf.resolve(config_node)
    return config_node


def cast(config: ConfigType):
    if is_dataclass(config):
        return from_dataclass(config)   # type: ignore
    elif isinstance(config, dict):
        return from_dict(config)
    elif isinstance(config, str):
        return from_string(config)
    elif isinstance(config, DictConfig):
        return config
    else:
        raise TypeError(f'Cannot cast `{type(config)}` to DictConfig')


def from_options(opts: Sequence[str]) -> DictConfig:
    config = OmegaConf.from_dotlist(list(opts))
    if not isinstance(config, DictConfig):
        raise TypeError(f"input is not a sequence of strings, but `{opts}")
    return config


def from_dataclass(config: Any) -> DictConfig:
    if not(is_dataclass(config)):
        msg = '`config_node` must be be decorated as a dataclass'
        raise ValueError(msg)
    config = OmegaConf.structured(config)
    if not isinstance(config, DictConfig):
        raise TypeError(f'Cannot create dict config from `{config}`')

    return config


def from_file(path: Optional[Path] = None) -> DictConfig:
    if path is not None:
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f'Cannot file configuration at {path}')
        config = OmegaConf.load(str(path))
    else:
        config = from_dict(None)

    if not isinstance(config, DictConfig):
        raise ValueError(f'Config loaded from {path} is not a DictConfig!')

    return config


def from_none() -> DictConfig:
    """Returns an empty dict config"""
    return OmegaConf.create()


def from_dict(
    config: Optional[Union[DictConfig, Dict[str, Any]]] = None
) -> DictConfig:
    if config is None:
        new_config = from_none()
    elif not isinstance(config, DictConfig):
        try:
            new_config = OmegaConf.create(config)
        except Exception as e:
            msg = f'Cannot get config from object of type `{type(config)}`'
            raise ValueError(msg) from e
    else:
        new_config = config

    if not isinstance(new_config, DictConfig):
        raise ValueError(f'Config `{config}` is not a DictConfig!')

    return new_config


def from_string(
    config: str,
    config_cls: Optional[Type[DataClass]] = None,
    strict: bool = False
) -> DictConfig:

    parsed_config = OmegaConf.create(config)
    if not isinstance(parsed_config, DictConfig):
        raise ValueError(f'Config `{config}` is not a DictConfig!')

    if config_cls is not None:
        base_config = OmegaConf.structured(config_cls)
        parsed_config = merge(base_config, parsed_config, strict=strict)

    if not isinstance(parsed_config, DictConfig):
        raise TypeError(f'Could not create config from string `{config}`')
    return parsed_config


def to_yaml(config: Union[DictConfig, Dict[str, Any]]) -> str:
    config = from_dict(config)
    return OmegaConf.to_yaml(config)


def to_dict(config: Union[DictConfig, Dict[str, Any]]) -> Dict[str, Any]:
    config = from_dict(config)
    return OmegaConf.to_container(config)


def register(
    name: str,
    use_cache: bool = False
) -> Callable[[Callable[..., RT]], Callable[..., RT]]:

    def _register(func: Callable[..., RT]) -> Callable[..., RT]:
        # will raise an error if the resolver is already registered
        OmegaConf.register_new_resolver(
            name=name, resolver=func, use_cache=use_cache, replace=False
        )
        return func

    return _register


@contextmanager
def editable(*configs: ConfigType, enable: bool) -> Iterator[bool]:
    try:
        with ExitStack() as stack:
            if enable:
                [stack.enter_context(open_dict(cast(c))) for c in configs]
        yield enable

    finally:
        pass


def merge(*configs: ConfigType, strict: bool = False) -> DictConfig:
    with editable(*configs, enable=not(strict)):
        merged_config = OmegaConf.merge(*configs)

    if not isinstance(merged_config, DictConfig):
        raise TypeError(f'Merged config {merged_config} is not a DictConfig!')

    return merged_config



from .cli import cli
from .core import (
    MISSING,

    cast,
    dataclass,
    editable,
    field,
    from_dict,
    from_file,
    from_none,
    from_options,
    from_string,
    merge,
    register,
    to_dict,
    to_yaml,
    traverse,
    validate
)
from .init import Target, init
from .logging import configure_logging


__all__ = [
    'cast',
    'cli',
    'configure_logging',
    'dataclass',
    'editable',
    'field',
    'from_dict',
    'from_file',
    'from_none',
    'from_options',
    'from_string',
    'init',
    'merge',
    'MISSING',
    'register',
    'Target',
    'to_dict',
    'to_yaml',
    'traverse',
    'validate',
]



import os
from argparse import ArgumentParser
from dataclasses import is_dataclass
from enum import Enum
from functools import partial
from inspect import getfile, getfullargspec, isclass
from pathlib import Path
from typing import Any, Callable, Optional, Sequence, Type

from omegaconf import DictConfig

from .core import (from_dataclass, from_file, from_options, merge, traverse,
                   validate)
from .init import InitLater
from .utils import PrintUtils, clean_multiline


class CliFlags(Enum):
    CONFIG = 'config'
    PARSED = 'parsed'
    QUIET = 'quiet'
    INPUTS = 'inputs'
    OPTIONS = 'options'
    DEBUG = 'debug'


def make_flags(opt_name: CliFlags) -> Sequence[str]:
    """Simple helper to create a list of flags for an
    option based on its name"""
    return f'-{opt_name.value[0]}', f'--{opt_name.value}'


class cli:
    @classmethod
    def _check_signature(cls: Type['cli'], func: Callable):
        expected_args = getfullargspec(func).args
        if len(expected_args) == 0:
            msg = (f'Function `{func.__name__}` cannot be decorated '
                   f'by `config_to_program` because it does '
                   f'not accept any argument.')
            raise RuntimeError(msg)
        elif len(expected_args) > 1:
            msg = (f'Function `{func.__name__}` cannot be decorated by '
                   f' `config_to_program` because it expects '
                   f'{len(expected_args)} > 1; If you want to pass extra '
                   f'arguments to this function, use kwargs with default '
                   f'values.')
            raise RuntimeError(msg)

    @classmethod
    def _check_args(cls: Type['cli'], func: Callable, args: Sequence[Any]):
        if len(args):
            msg = (f'After decorating `{func.__name__}` with '
                   f'`config_to_program`, do not provide any additional '
                   f'arguments while invoking it; any additional parameter '
                   f'should be passed as a keyword argument.')
            raise RuntimeError(msg)

    @classmethod
    def _make_argument_parser(
        cls: Type['cli'],
        func: Callable,
        name: str
    ) -> ArgumentParser:
        """Sets up argument parser ahead of running the CLI. This includes
        creating a help message, and adding a series of flags."""

        # we find the path to the script we are decorating with the
        # cli so that we can display that to the user.
        current_dir = Path(os.getcwd())
        path_to_fn_file = Path(getfile(func))
        rel_fn_file_path = str(path_to_fn_file).replace(str(current_dir), '')

        # Program name and usage printed here.
        prog = f'Parser for configuration {name}'
        usage = clean_multiline(f'''
            python3 {rel_fn_file_path} '
            {{{"/".join(make_flags(CliFlags.CONFIG))} config_file.yaml}}
            {{{"/".join(make_flags(CliFlags.OPTIONS))}}}
            {{{"/".join(make_flags(CliFlags.INPUTS))}}}
            {{{"/".join(make_flags(CliFlags.DEBUG))}}}
            {{{"/".join(make_flags(CliFlags.PARSED))}}}
            {{{"/".join(make_flags(CliFlags.QUIET))}}}
            param1=value1, …, paramN=valueN'
        ''')
        ap = ArgumentParser(prog=prog, usage=usage)

        # add options
        msg = 'A path to a YAML file containing a configuration.'
        ap.add_argument(*make_flags(CliFlags.CONFIG),
                        default=None,
                        help=msg,
                        metavar='/path/to/config.yaml')

        msg = 'Print all default options and CLI flags.'
        ap.add_argument(*make_flags(CliFlags.OPTIONS),
                        action='store_true',
                        help=msg)

        msg = 'Print the input configuration.'
        ap.add_argument(*make_flags(CliFlags.INPUTS),
                        action='store_true',
                        help=msg)

        msg = 'Print the parsed configuration.'
        ap.add_argument(*make_flags(CliFlags.PARSED),
                        action='store_true',
                        help=msg)

        msg = 'Enter debug mode by setting global logging to DEBUG.'
        ap.add_argument(*make_flags(CliFlags.DEBUG),
                        action='store_true',
                        help=msg)

        msg = 'If provided, it does not print the configuration when running.'
        ap.add_argument(*make_flags(CliFlags.QUIET),
                        action='store_true',
                        help=msg)
        return ap

    @classmethod
    def _wrapped_main_method(
        cls: Type['cli'],
        func: Callable,
        name: str,
        config_node: DictConfig,
        print_fn: Optional[Callable] = None,
        strict_input: bool = False,
        *args: Any,
        **kwargs: Any
    ) -> Callable:

        if not isinstance(config_node, DictConfig):
            raise TypeError("Config node must be a DictConfig")

        # Making sure I can decorate this function
        cls._check_signature(func=func)
        cls._check_args(func=func, args=args)

        # Get argument parser and arguments
        ap = cls._make_argument_parser(func=func, name=name)
        opts, leftover_args = ap.parse_known_args()

        # set some default options for when no options are provided
        # printing_steps = PrintingSteps(opts)

        # setup debug
        if opts.debug:
            # relative import here not to mess things up
            from .logging import configure_logging
            configure_logging.debug()

        # Setup an utility to deal with printing
        pu = PrintUtils(print_fn=print_fn)

        # We don't run the main program if the user
        # has requested to print the any of the config.
        do_no_run = (opts.options or opts.inputs or opts.parsed)

        # Print default options if requested py the user
        if opts.options:
            params = traverse(config_node)

            cli_opts_repr = ('OPTS/CLI FLAG:', ) + tuple(
                f'{p.path} = ' + (str(p.value) if p.value != '' else "''")
                for p in params
            )
            pu.print(*cli_opts_repr, level_up=1)

        # load cli config and file config;
        # when merging, cli_config takes precedence over file_config
        cli_config = from_options(leftover_args)
        file_config = from_file(opts.config)
        input_config = merge(file_config, cli_config)

        # print both configs if requested
        if opts.inputs:
            pu.print('INPUT/CLI ARGS:', cli_config)
            pu.print('INPUT/CFG FILE:', file_config)

        if do_no_run and not opts.parsed:
            # if the user hasn't requested to print the parsed config
            # and we are not running the main program, we can exit here.
            return InitLater.no_op()

        # load configuration with node parsers
        parsed_config = merge(config_node, input_config, strict=strict_input)

        # check if all parameters are provided/resolved
        parsed_config = validate(parsed_config)

        # print it if requested
        if not(opts.quiet) or opts.parsed:
            pu.print('PARSE/ALL CFG:', parsed_config)   # type: ignore

        if do_no_run:
            # we are not running because the user has requested to print
            # either the options, inputs, or parsed config.
            return InitLater.no_op()
        else:
            # we execute the main method and pass the parsed config to it
            return func(parsed_config, **kwargs)

    def __new__(cls: Type['cli'],
                config_node_cls: Any,
                strict_input: bool = False,
                print_fn: Optional[Callable] = None) -> Callable:

        if not(isclass(config_node_cls) and is_dataclass(config_node_cls)):
            msg = '`config_node` must be be decorated as a dataclass'
            raise ValueError(msg)

        config_node = from_dataclass(config_node_cls)
        name = config_node_cls.__name__

        def wrapper(func: Callable) -> Callable:
            return partial(cls._wrapped_main_method,
                           func=func,
                           name=name,
                           strict_input=strict_input,
                           config_node=config_node,
                           print_fn=print_fn)
        return wrapper



import functools
import importlib
import importlib.util
import inspect
import itertools
from typing import Any, Callable, Dict, Optional, Type, Union

from omegaconf import DictConfig, OmegaConf

from .core import from_dict
from .utils import clean_multiline


class InitLater(functools.partial):
    def get_kw(self: 'InitLater', *args, **kwargs):
        """Shortcut for accessing parameters that have been
        provided to an InitLater object"""
        return self.keywords.get(*args, **kwargs)

    def pop_kw(self: 'InitLater', *args, **kwargs):
        """Shortcut for popping parameters that have been
        provided to an InitLater object"""
        return self.keywords.pop(*args, **kwargs)

    @staticmethod
    def _no_op():
        ...

    @classmethod
    def no_op(cls: Type['InitLater']) -> 'InitLater':
        """Create an init later that does nothing.
        Useful for when trying to instantiate from None."""
        return cls(cls._no_op)

    def __bool__(self: 'InitLater') -> bool:
        # this allows us to check if the InitLater object
        # is "True" (that is, it is a real wrapped function),
        # or if it is "False" (it's a no-op)
        return self.func != self._no_op

    def __call__(self: 'InitLater', /, *args, **kwargs):
        # recursively call deferred initialization if
        # we encounter another InitLater
        args = [v() if isinstance(v, InitLater) else v
                for v in itertools.chain(self.args, args)]
        kwargs = {k: v() if isinstance(v, InitLater) else v
                  for k, v in {**self.keywords, **kwargs}.items()}
        try:
            return self.func(*args, **kwargs)
        except Exception as e:
            ex_name = type(e).__name__
            fn_name = repr(self.func)
            msg = clean_multiline(f'''
                An error occurred while trying to initialize {fn_name}
                with arguments "{args}" and kwargs "{kwargs}":
                {ex_name}("{' '.join(map(str, e.args))}")
            ''')
            raise RuntimeError(msg).with_traceback(e.__traceback__)


class Target:
    @staticmethod
    def _is_module(path: str) -> bool:
        try:
            spec = importlib.util.find_spec(path)
            return spec is not None
        except ModuleNotFoundError:
            return False

    @classmethod
    def to_string(cls: Type['Target'], callable: Any) -> str:
        if inspect.isclass(callable):
            return f'{callable.__module__}.{callable.__name__}'
        elif inspect.ismethod(callable):
            return f'{cls.to_string(callable.__self__)}.{callable.__name__}'
        elif inspect.isfunction(callable):
            return f'{callable.__module__}.{callable.__name__}'
        else:
            raise TypeError(f'`{callable}` is not a callable')

    @classmethod
    def from_string(cls: Type['Target'], path: str) -> Any:
        if cls._is_module(path):
            return importlib.import_module(path)
        elif '.' in path:
            m_name, c_name = path.rsplit('.', 1)
            if cls._is_module(m_name):
                container = importlib.import_module(m_name)
            else:
                container = cls.from_string(m_name)
            callable_ = getattr(container, c_name, None)
        else:
            callable_ = globals().get(path, __builtins__.get(path, None))

        if callable_ is None:
            raise ImportError(f'Cannot find callable at {path}')

        return callable_


class init:
    TARGET: str = '_target_'

    @classmethod
    def callable(cls: Type['init'],
                 config: Union[DictConfig, Dict[str, Any]]) -> Callable:

        config = from_dict(config)

        try:
            target: str = config[cls.TARGET]
        except KeyError:
            raise KeyError(f'Config `{config}` has no `{cls.TARGET}` key!')

        return Target.from_string(target)

    @classmethod
    def later(
        cls: Type['init'],
        config: Optional[Union[Dict[str, Any], DictConfig]] = None,
        _recursive_: bool = True,
        **kwargs: Dict[str, Any]
    ) -> InitLater:
        """Return a InitLater object to be used to instantiate a
        new class or call a function"""

        # if no config is provided, we return a function
        if config is None:
            return InitLater.no_op()

        config_node = from_dict(config)

        if len(kwargs) > 0:
            config_node = OmegaConf.merge(
                config_node, from_dict(kwargs)
            )

        if cls.TARGET not in config_node:
            msg = (f'Cannot instantiate from `{config_node}`: '
                   f'`{cls.TARGET}` keyword missing')
            raise ValueError(msg)

        fn = cls.callable(config_node)      # type: ignore

        def _recursive_init(param):
            must_recursive_init = (_recursive_ and
                                   isinstance(param, (DictConfig, dict))
                                   and cls.TARGET in param)
            if must_recursive_init:
                param = cls.later(config=param, _recursive_=True)
            return param

        init_call_dict = {k: _recursive_init(v)
                          for k, v in config_node.items()
                          if k != cls.TARGET}

        return InitLater(fn, **init_call_dict)

    @classmethod
    def now(
        cls: Type['init'],
        config: Optional[Union[DictConfig, dict]] = None,
        _recursive_: bool = True,
        **kwargs: Dict[str, Any]
    ) -> object:
        """Create a later, but initialize it now!"""
        if config is None:
            return None

        return cls.later(config=config, _recursive_=_recursive_, **kwargs)()



import logging
import warnings
from multiprocessing import current_process
from pathlib import Path
from typing import Optional, Type, Union


class configure_logging:
    _DEBUG_MODE = None

    @classmethod
    def debug(cls: Type['configure_logging'],
              *args,
              logging_level: Optional[int] = None,
              **kwargs) -> logging.Logger:

        if logging_level is not None:
            msg = '`logging_level` was provided to `debug`, but it is ignored'
            warnings.warn(msg)

        cls._DEBUG_MODE = True
        return cls(*args, logging_level=logging.DEBUG, **kwargs)

    def __new__(
        cls: Type['configure_logging'],
        logger_name: Optional[str] = None,
        file_logging_path: Optional[Path] = None,
        make_dir_if_missing: bool = True,
        fmt: str = "[%(asctime)s][%(name)s][%(levelname)s] %(message)s",
        datefmt: str = "%Y-%m-%d %H:%M:%S",
        logging_level: Union[int, str] = logging.INFO,
        force_root_reattach: bool = True,
        root_formatter_kwargs: Optional[dict] = None,
        stream_handler_kwargs: Optional[dict] = None,
        file_handler_kwargs: Optional[dict] = None,
        basic_config_kwargs: Optional[dict] = None,
        additional_handlers: Optional[list] = None,
    ) -> logging.Logger:
        """A big function that speeds up configuration of logging"""

        if isinstance(logging_level, str):
            logging_level = getattr(logging, logging_level)

        # we abide by the global debugging flag
        logging_level = logging.DEBUG if cls._DEBUG_MODE else logging_level

        # change how the formatter looks
        kw = root_formatter_kwargs or {}
        root_formatter = logging.Formatter(fmt=fmt, datefmt=datefmt, **kw)

        handlers = list(additional_handlers or [])

        kw = stream_handler_kwargs or {}
        root_stream_handler = logging.StreamHandler(**kw)
        root_stream_handler.setFormatter(root_formatter)
        handlers.append(root_stream_handler)

        if file_logging_path is not None:
            file_logging_path = Path(file_logging_path)

            if make_dir_if_missing:
                file_logging_path.parent.mkdir(parents=True, exist_ok=True)

            path_name = (f'{file_logging_path.stem}'
                         f'_{current_process().name}'
                         f'{file_logging_path.suffix}')

            kw = file_handler_kwargs or {}
            root_file_handler = logging.FileHandler(
                filename=file_logging_path.parent / path_name,
                **kw
            )
            root_file_handler.setFormatter(root_formatter)
            handlers.append(root_file_handler)

        # add the handlers to the root logger, force reattaching other loggers
        kw = basic_config_kwargs or {}
        logging.basicConfig(level=logging_level,
                            force=force_root_reattach,
                            handlers=handlers, **kw)
        return logging.getLogger(logger_name)



import re
import shutil
from dataclasses import dataclass, field
from textwrap import dedent
from typing import Callable, Optional, Union

import yaml
from omegaconf import DictConfig, OmegaConf


def clean_multiline(string: str) -> str:
    string = dedent(string.strip())
    string = re.sub(r'\s*\n\s*', ' ', string)
    return string


@dataclass
class PrintUtils:
    """A few utilities to make printing easier"""

    indent_step: int = 2
    indent_char: str = ' '
    separator_char: str = '-'
    terminal_width: int = field(default=shutil.get_terminal_size().columns)
    print_fn: Optional[Callable] = field(default_factory=lambda: print)

    def __post_init__(self):
        self.print_fn = self.print_fn or print
        self.has_printed_first_separator = False
        self.is_default_print = self.print_fn is print

    def print(self,
              *lines: Union[str, dict, DictConfig],
              level: int = 0,
              level_up: Optional[int] = None):

        level_up = level_up or float('inf')     # type: ignore

        content = ''
        if self.is_default_print:
            if not self.has_printed_first_separator:
                content += self.separator()
            self.has_printed_first_separator = True

        for line in lines:
            if not line:
                # backtrack the new line you just added from previous
                # argument if the current argument evaluates to none, like
                # an empty string, or an empty dict
                content = f'{content.rstrip()} '

            if isinstance(line, DictConfig):
                line = OmegaConf.to_container(line)

            if isinstance(line, dict):
                content += f'{self.to_yaml(line, level)}\n'
            else:
                content += f'{self.indent(line, level)}\n'

            level += 1 if level < level_up else 0   # type: ignore

        if self.is_default_print:
            content += f'{self.separator()}\n'

        self.print_fn(content.rstrip())     # type: ignore

    def indent(self, line: str, level: int) -> str:
        indent = self.indent_char * level * self.indent_step
        return indent + line

    def separator(self, level: int = 0) -> str:
        separator = self.separator_char * self.terminal_width
        return self.indent(separator, level=level)

    def to_yaml(self, content: dict, level: int = 0) -> str:
        if not content:
            # this is in case the dict/list is empty
            return str(content)

        out = yaml.dump(content,
                        indent=self.indent_step,
                        width=self.terminal_width,
                        default_flow_style=False)
        out = '\n'.join(self.indent(ln, level)
                        for ln in out.strip().split('\n')).rstrip()
        return out
